"""
ListSorters used to control if and how Lists are sorted.  Really there are only two
choices.  Sort or don't sort.  The sorter can be applied to the List elements by
the Selector that is associated with the sorter.
"""
from abc import abstractmethod
from pathlib import Path
from typing import List, Optional, Union

from .selector import BaseSelector, SELECTORS


class NotSortedError(Exception):
    """The sorter was not applied to the list."""


class BaseListSorter:
    """
    Base list sorter implements the chaining logic.
    """

    def __init__(
        self, selectors: SELECTORS = None,
    ):
        """
        Initialize the sorter with optional parent sorter and selector.

        :param selectors: Optional selectors to use to select which
            elements this sort runs.
        """
        if selectors:
            self._selectors = selectors if isinstance(selectors, list) else [selectors]
        else:
            # No selectors specified
            self._selectors = None

    @staticmethod
    def sorted(
        list_: List, path: Path, sorters: Optional[List["BaseListSorter"]] = None
    ) -> List:
        """
        Run all the sorters until one applied to sort or not sort the list.

        :param list_: List to sort.
        :param path: Path to the element.
        :param sorters: List of sorters to use to sort the lists.
        :return: Sorted list.
        """
        if not sorters:
            return sorted(list_)

        for sorter in sorters:
            if BaseSelector.match(path, sorter._selectors):
                try:
                    return sorter._sorted(list_)
                except NotSortedError:
                    continue
        return sorted(list_)

    @abstractmethod
    def _sorted(self, list_: List) -> List:
        """
        Prototype for the core sorting logic implemented in the subclass.

        :param list_: List to sort.
        :return: Sorted list.
        """
        pass  # pragma: no cover


LIST_SORTERS = Optional[Union[BaseListSorter, List[BaseListSorter]]]


class DefaultListSorter(BaseListSorter):
    def __init__(
        self, *, selectors: SELECTORS = None,
    ):
        """
        Standard Python sorted() sorter.

        :param selectors: Optional list of selectors to use to select which
            elements this sort runs.
        """
        super().__init__(selectors)

    def _sorted(self, list_: List) -> List:
        """Default Python sorted()."""
        return sorted(list_)


class NoSortListSorter(BaseListSorter):
    def __init__(
        self, *, selectors: SELECTORS = None,
    ):
        """
        No Op sorter.

        :param selectors: Optional list of selectors to use to select which
            elements this sort runs.
        """
        super().__init__(selectors)

    def _sorted(self, list_: List) -> List:
        """No Op sort."""
        return list_



"""
Common Selectors used to determine if a Normalizer or ListSorter will
be run on a leaf element or an List.  These can be extened to create any
number of other methods for determining if the path is matched.

Selectorss can be chained so that if the first selector doesn't match successive calls
to parent selectors will be made until a match is found or all selectors have been exhausted.
"""
import re
from abc import abstractmethod
from pathlib import Path
from typing import List, Optional, Union


class BaseSelector:
    """
    Base path selector implements the chaining logic.
    """

    def __init__(self):
        """
        Initialize the selector.
        """

    @staticmethod
    def match(path: Path, selectors: Optional[List["BaseSelector"]] = None) -> bool:
        """
        Match the given path against the chain of selectors.

        :param path: Path to match.
        :param selectors: List of selectors.
        :return: True if matched.
        """
        if not selectors:
            return True

        for selector in selectors:
            if selector._match(path):
                return True
        return False

    @abstractmethod
    def _match(self, path: Path) -> bool:
        """
        Prototype for the core matching logic of any PathSelectors subclass.

        :param path: Path to match.
        :return: True if matched.
        """
        pass  # pragma: no cover


SELECTORS = Optional[Union[BaseSelector, List[BaseSelector]]]


class ListLastComponentSelector(BaseSelector):
    """
    Match the last component of the path against a list strings.
    """

    def __init__(self, component_names: List):
        """
        Selector with list of last component names to match.

        :param component_names: List of component names to match.
        """
        self._component_names = component_names
        super().__init__()

    def _match(self, path: Path) -> bool:
        """
        Match the path's last component against the list of match strings.

        :param path:  Path to match.
        :return: True if matched.
        """
        return path.parts[-1] in self._component_names


class ListAnyComponentSelector(BaseSelector):
    """
    Match any component in the path against a list of string.
    """

    def __init__(self, component_names: List):
        """
        Selectors that matches any component of the list against a list of match strings.

        :param component_names: List of component names to match.
        """
        self._component_names = component_names
        super().__init__()

    def _match(self, path: Path) -> bool:
        """
        Match the path's components against the list of match strings.

        :param path: Path to match.
        :return: True if matched.
        """
        return any((part in self._component_names for part in path.parts))


class RegExSelector(BaseSelector):
    """
    Match the regex against the path.
    """

    def __init__(self, regex: str):
        """
        Selectors that matches the path with a RegEx.

        :param regex: Regex to use to search the path.
        """
        self._regex = re.compile(regex)
        super().__init__()

    def _match(self, path: Path) -> bool:
        """
        Match the path with a search for the RegEx.

        :param path: Path to match.
        :return: True if matched.
        """
        return self._regex.search(str(path)) is not None


class NegativeSelector(BaseSelector):
    """
    Selector that inverts another selectors match results.
    """

    def __init__(self, selector: BaseSelector):
        """
        Negate the match of the child path selector.

        :param selector: Child path selector to negate.
        """
        self._selector = selector
        super().__init__()

    def _match(self, path: Path) -> bool:
        """
        Negate the match result of the child.

        :param path: Path to match.
        :return: True if the path wasn't matched.
        """
        return not self._selector.match(path, [self._selector])



from .differ import DiffResult, Differ
from .list_sorter import BaseListSorter, NoSortListSorter, DefaultListSorter
from .normalizer import (
    NORMALIZERS,
    BaseNormalizer,
    DefaultNormalizer,
    FloatRoundNormalizer,
    TodayDateNormalizer,
    StrTodayDateNormalizer,
)
from .selector import (
    SELECTORS,
    BaseSelector,
    ListLastComponentSelector,
    ListAnyComponentSelector,
    RegExSelector,
    NegativeSelector,
)
from .sorter import Sorter



"""
Compare to nested dictionary/list objects.  diff() will return a unix diff like
list of lines of the jsonified object to help locate the differences.
"""
import json
from difflib import HtmlDiff
from json import JSONEncoder
from typing import List, Optional, Type

from .formatter import Formatter
from .list_sorter import LIST_SORTERS
from .normalizer import NORMALIZERS
from .sorter import Sorter, NDLElement


class DiffResult:
    """
    Result of a compare or diff.   Acts like a bool for testing purposes.
    Provides supporting information for the match.
    """

    def __init__(self, match: bool, support: List[str]):
        self._match = match
        self.support = support

    def __bool__(self) -> bool:
        return self._match


class Differ:
    """
    Provides comparision and difference methods for two objects of
    nested dictionary/lists.   The process is to first sort the two objects
    and then compare them or jsonify them and compare the individual lines.
    """

    @staticmethod
    def diff(
        left: NDLElement,
        right: NDLElement,
        cls: Optional[Type[JSONEncoder]] = None,
        sorters: LIST_SORTERS = None,
        normalizers: NORMALIZERS = None,
        max_col_width: Optional[int] = 20,
    ) -> DiffResult:
        """
        Show the difference of two objects.  Unix like diff results.
        :param left: Test object
        :param right: Expected object
        :param cls: JSON Encoder if any fields aren't JSON encodable.
        :param sorters: Sorters for list elements.
        :param normalizers: Normalizers for leaf elements.
        :param max_col_width: Maximum column width of diff output.
        :return: True if match.
        """
        if normalizers:
            normalizers = (
                normalizers if isinstance(normalizers, list) else [normalizers]
            )
        sorted_left = Sorter.sorted(left, sorters=sorters, normalizers=normalizers)
        sorted_right = Sorter.sorted(right, sorters=sorters, normalizers=normalizers)
        differ = HtmlDiff()

        result = differ.make_file(
            json.dumps(sorted_left, indent=2, cls=cls).split("\n"),
            json.dumps(sorted_right, indent=2, cls=cls).split("\n"),
        )
        match, support = Formatter(max_col_width=max_col_width).format(result)
        return DiffResult(match, support)



"""
Sort nested dictionary/lists.  The sort order isn't really important,
just that it is consistent.

Alternative ListSorters can be applied to elements selected by the
by Selectors.
"""
from pathlib import Path
from typing import Any, Union, Mapping, Optional, List, Dict

from .list_sorter import BaseListSorter, LIST_SORTERS
from .normalizer import BaseNormalizer, NORMALIZERS

NDLElement = Union[Mapping, List, Any]


class SortedMapping(dict):
    """
    Replacement for a dictionary that sorts it's keys when it is created.
    Does depth first replacement of dicts and lists so that when it sorts
    it is using the sorted version of its contents.
    """

    def __init__(
        self,
        data: Mapping,
        path: Path,
        sorters: Optional[List[BaseListSorter]] = None,
        normalizers: Optional[List[BaseNormalizer]] = None,
    ):
        """
        Construct a new dict that is sorted.
        :param data: Unsorted dictionary.
        :param path: Path to current element.
        :param sorters: Sorters for list elements.
        :param normalizers: Normalizers for leaf nodes.
        """
        super().__init__(
            **{
                k: Sorter._sorted(data[k], path / k, sorters, normalizers)
                for k in sorted(data.keys())
            }
        )

    def __lt__(self, other) -> bool:
        """
        Compare two objects.  If they are both SortedMapping then compare them
        as dictionaries.  Otherwise, just compare their class types.  Order isn't
        really important.  It just needs to be consistent.
        :param other: Object to compare.
        :return: bool
        """
        if isinstance(other, SortedMapping):
            return list(self.items()) < list(other.items())
        else:
            # The order doesn't really matter here as long as it is consistent.
            return str(self.__class__) < str(other.__class__)


class SortedList(list):
    """
    Replacement for a list that sorts it's keys when it is created.
    Does depth first replacement of dicts and lists so that when it sorts
    it is using the sorted version of its contents.
    """

    def __init__(
        self,
        list_: List,
        path: Path,
        sorters: Optional[List[BaseListSorter]] = None,
        normalizers: Optional[List[BaseNormalizer]] = None,
    ):
        """
        Construct a new list that is sorted.
        :param list_: Unsorted list.
        :param path: Path to the current element.
        :param sorters: Sorters for list elements.
        :param normalizers: Normalizers for leaf elements.
        """
        sorted_children = [
            Sorter._sorted(v, path / f"[{i}]", sorters, normalizers)
            for i, v in enumerate(list_)
        ]
        super().__init__(BaseListSorter.sorted(sorted_children, path, sorters))

    def __lt__(self, other) -> bool:
        """
        Compare two objects.  If they are both SortedList then compare them
        as lists.  Otherwise, just compare their class types.  Order isn't
        really important.  It just needs to be consistent.
        :param other: Object to compare.
        :return: bool
        """
        if isinstance(other, SortedList):
            return list(self) < list(other)
        else:
            # The order doesn't really matter here as long as it is consistent.
            return str(self.__class__) < str(other.__class__)


class Sorter:
    @staticmethod
    def _sorted(
        data: NDLElement,
        path: Path,
        sorters: Optional[List[BaseListSorter]] = None,
        normalizers: Optional[List[BaseNormalizer]] = None,
    ) -> Union[SortedMapping, SortedList, Any]:
        """
        Sort a nested dictionary/list.  Used internally to keep the path argument from being exposed in the
        public interface.
        :param data: Object to sort.
        :param path: Path to the current element.
        :param sorter: Sorter for list elements.
        :param normalizers: List of normalizer for leaf elements.
        :return: Sorted object.
        """
        if isinstance(data, Dict):
            return SortedMapping(data, path, sorters, normalizers)
        elif isinstance(data, List):
            return SortedList(data, path, sorters, normalizers)
        else:
            return BaseNormalizer.normalize(data, path, normalizers)

    @staticmethod
    def sorted(
        data: NDLElement,
        *,
        sorters: LIST_SORTERS = None,
        normalizers: NORMALIZERS = None,
    ) -> Union[SortedMapping, SortedList, Any]:
        """
        Sort a nested dictionary/list.
        :param data: Object to sort.
        :param sorters: Sorter for list elements.
        :param normalizers: Normalizer for leaf elements.
        :return: Sorted object.
        """
        if sorters:
            sorters = sorters if isinstance(sorters, list) else [sorters]
        if normalizers:
            normalizers = (
                normalizers if isinstance(normalizers, list) else [normalizers]
            )

        return Sorter._sorted(data, Path(), sorters=sorters, normalizers=normalizers)



"""
Quick hack to see what it will take to format the HTML output from the difflib.html_diff() into
plain text that can be displayed in pytest.
"""
from html.parser import HTMLParser
from pathlib import Path
from typing import Optional

ADD_FORMAT_ON = "\033[0;32m"
SUB_FORMAT_ON = "\033[0:31m"
CHANGE_FORMAT_ON = "\033[0:34m"
START_MARKS = {
    "diff_add": ADD_FORMAT_ON,
    "diff_sub": SUB_FORMAT_ON,
    "diff_chg": CHANGE_FORMAT_ON,
}
FORMAT_OFF = "\033[0m"
FORMAT_EXTRA_CHARS = len(ADD_FORMAT_ON) + len(FORMAT_OFF)

# ToDo: data really needs to be passed as a list.  Then the process of picking off enough characters
#       to format to the correct width will be easier.   It will also allow for the correct
#       format off character to be added if needed.


class Row:
    """
    Collect up each of the columns in a list and then format them all together at the end.
    """

    def __init__(self, max_col_width: int):
        """
        Start a new row of output.

        :param max_col_width: Future support for truncating or smart selection of a portion of a line.
        """
        self.max_col_width = max_col_width
        self.data = list()

    def add(self, data: str):
        """Add the next column to the row."""
        self.data.append(data.replace("\xa0", " ") if data else data)

    def finalize(self):
        """Concatenate the line together with any coloring required.  Truncate to max columns."""
        left = self.data[2] if self.data[2] else ""
        right = self.data[5] if self.data[5] else ""

        left_chars = (
            self.max_col_width + FORMAT_EXTRA_CHARS
            if FORMAT_OFF in left
            else self.max_col_width
        )
        right_chars = (
            self.max_col_width + FORMAT_EXTRA_CHARS
            if FORMAT_OFF in right
            else self.max_col_width
        )

        return f"{left:{left_chars}} {right:{right_chars}}"


class Formatter(HTMLParser):
    def __init__(self, max_col_width: Optional[int] = 20):
        """
        Keep track of all the state variables for parsing each of the rows.

        :param max_col_width: Limit for how wide any line can be.
        """
        self.max_col_width = max_col_width
        self.in_table = False
        self.done = False
        self.row = None
        self.looking_for_data = False
        self.data = None
        self.change_mark = None
        self.output = list()
        self.match = True
        super().__init__()

    def handle_starttag(self, tag, attrs):
        """
        Process the start tags to change the state of the parser.
        """
        if tag == "tbody" and not self.done:
            self.in_table = True
            return
        if self.in_table and tag == "tr":
            self.row = Row(self.max_col_width)
        if self.in_table and tag == "td":
            self.looking_for_data = True
            self.data = None
        if self.looking_for_data and tag == "span":
            class_attr = [value for attr, value in attrs if attr == "class"][0]
            self.match = False
            self.change_mark = START_MARKS[class_attr]
            self.data = (
                "".join([self.data, self.change_mark])
                if self.data
                else self.change_mark
            )

    def handle_endtag(self, tag):
        """
        Process the end tags to change the state of the parser.
        """
        if tag == "tbody":
            self.in_table = False
            return
        if self.in_table and tag == "tr":
            self.output.append(self.row.finalize())
            self.row = None
        if self.in_table and tag == "td":
            self.row.add(self.data)
            self.looking_for_data = False
            self.data = None
        if self.looking_for_data and tag == "span":
            self.data = (
                "".join([self.data, FORMAT_OFF]) if self.data else self.change_mark
            )

    def handle_data(self, data):
        if self.looking_for_data:
            self.data = "".join([self.data, data]) if self.data else data

    def format(self, diff: str):
        """Parse and format the html into a colored test format."""
        self.feed(diff)
        return self.match, "\n".join(self.output)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ndl_tools']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'ndl-tools',
    'version': '0.0.12',
    'description': 'Tools for sorting and diffing nested dictionaries and lists.',
    'long_description': '# ndl-tools\n![test](https://github.com/nathan5280/ndl-tools/workflows/test/badge.svg)\n\nTools for sorting and diffing nested dictionaries and lists.  \n\nThe focus of the package is to support API testing.  Hashing two object trees or \nnested dictionaries to compare them works great when they are \nactually equal. If they aren\'t equal finding out why can become quite tedious.\n\nAre these equal?\n```python\nobj1 = {"a": 1, "b": 2}\nobj2 = {"b": 2, "a": 1}\n```\n\nWhat about a set that is mapped into a list when converted to JSON.\n```python\nobj1 = [1, 2, 3, 4]\nobj2 = [4, 3, 2, 1]\n```\n\n```python\nfrom datetime import datetime\n\nobj1 = {"start_date": datetime.date(1999, 1, 1)}\nobj2 = {"start_date": datetime.date(2020, 8, 19)}\n```\n\nThe dictionary isn\'t to bad to get sorted and compared correctly, but it gets messy when one \nof the values is another dictionary, list or set.  The list case needs to be sorted or not sorted\ndepending on the context of the object and if it is a list or a set.  And finally, the date one\nis hard to keep up to date in your test cases because the dates keep shifting. It isn\'t to bad if you \ncan use something like freezegun to go back in time, but if the payload comes from an external \nservice it can be a mess.\n\n## Concepts\n| Term | Definition |\n| :--- | :--- |\n| Differ | Entry point to support diff of two Nested-Dict-Lists (NDL). |\n| DiffResult | Result of calling diff() in a Differ object.   It acts like a bool for simple asserts, but also provides a two column colored difference of the two NDL. |\n| ListSorter | Classes that can be selectively applied to lists in the NDL either sort or not sort a list. |\n| Normalizer | Classes that can be applied to leaf elements to transform them to match from the left and right NDLs. |\n| Selector | Classes used to select what elements in the traversal of the NDL a given ListSorter or Normalizer is applied. |\n| Sorter | Entry point to NDL sorter functionality. This normally isn\'t used directly as it sits behind the Differ. |\n\n## Examples\n### Float Precision\n```python\nfrom ndl_tools import Differ\n\nOBJ_1 = {"a": 1.0, "b": 2.01}\nOBJ_2 = {"a": 1.01, "b": 2.011}\n\n\ndef float_mismatch():\n    differ = Differ()\n    result = differ.diff(OBJ_1, OBJ_2)\n    assert not result\n    print(result.support)\n```\n<img src="https://github.com/nathan5280/ndl-tools/blob/develop/images/float-precision-fail.png" height="75"/>\n\nNote the highlights on the differences.  Red will indicate that something was deleted and blue that\nsomething was changed and yellow that something was added.\n\n#### Match\nLets apply the *FloatRoundNormalizer* when we do the diff and see if we can get the NDLs to match.\n\n```python\nfrom ndl_tools import Differ, FloatRoundNormalizer\n\nOBJ_1 = {"a": 1.0, "b": 2.01}\nOBJ_2 = {"a": 1.01, "b": 2.011}\n\n\ndef float_match():\n    differ = Differ()\n    float_round_normalizer = FloatRoundNormalizer(places=1)\n    result = differ.diff(OBJ_1, OBJ_2, normalizers=[float_round_normalizer])\n    assert result\n    print(result.support)\n```\n\n<img src="https://github.com/nathan5280/ndl-tools/blob/develop/images/float-precision-pass.png" height="75"/>\n\n\n#### Selector to Apply Different Nomalizers\n```python\nfrom ndl_tools import Differ, FloatRoundNormalizer, ListLastComponentSelector\n\nOBJ_1 = {"a": 1.0, "b": 2.01}\nOBJ_2 = {"a": 1.01, "b": 2.011}\n\n\ndef float_two_precision_match():\n    differ = Differ()\n    # Normalize the \'a\' element to 1 decimal place.\n    a_selector = ListLastComponentSelector(component_names=["a"])\n    one_float_round_normalizer = FloatRoundNormalizer(places=1, selectors=[a_selector])\n\n    # Normalize the \'b\' element to 2 decimal places.\n    b_selector = ListLastComponentSelector(component_names=["b"])\n    two_float_round_normalizer = FloatRoundNormalizer(\n        places=2, selectors=[b_selector]\n    )\n\n    result = differ.diff(OBJ_1, OBJ_2, normalizers=[two_float_round_normalizer, one_float_round_normalizer])\n    assert result\n    print(result.support)\n```\n\n<img src="https://github.com/nathan5280/ndl-tools/blob/develop/images/float-two-precision-pass.png" height="75"/>\n\nEach of the Normalizers can have a different selector or use the default which is to apply it to\nall elements.  The list of Normalizers are called in order until one normalizes the element or all \nnormalizers are exhausted.  There is an art to figuring out how to minimize the number of \nNormalizers and  Selectors you need to get two NDLs to match.   If you start getting to \nthe point where you have many of them it might be time to think about doing some \nprework on the NDL before comparing them.\n\n# Normalizers\nNormalizers are designed to be easily extensible.  Checkout the existing [Normalizers](https://github.com/nathan5280/ndl-tools/blob/develop/src/ndl_tools/normalizer.py)\nYou can easily see ways to extend these to support exponential numbers, dates, ...\n\n| Normalizer | Usage |\n| :--- | :---|\n| FloatRoundNormalizer | Round a floating point number to a set number of places. |\n| TodayDateNormalizer | Set the date to datetime.date.today(). |\n| StrTodayDateNormalizer | Convert a string representation of a date to string representation of today.  Useful if one of the NDLs was read from JSON and the dates weren\'t converted. |\n\nHave some fun building your own Normalizers.   It only takes a few lines in the __init__() and _normalize() methods.\n\n>[!WARNING]\n>If a normalizer was applied to an element, but doesn\'t actually normalize it, the normalizer should raise NotNormalizedError()\n\n# Selectors\nSelectors determine if the normalizer they are attached to will be applied to a given element.  Again \nthere is an art to figuring out the minimum number needed or the minimum that are still clear. \n\nWhile this isn\'t the most efficient way to rewrite the example above that rounds both \'a\' and \'b\' to \none decimal place, it does show how multiple selectors can be applied to a single normalizer.\n\n```python\nfrom ndl_tools import Differ, FloatRoundNormalizer, ListLastComponentSelector\n\nOBJ_1 = {"a": 1.0, "b": 2.01}\nOBJ_2 = {"a": 1.01, "b": 2.011}\n\n\ndef selector_chaining_match():\n    differ = Differ()\n\n    a_selector = ListLastComponentSelector(component_names=["a"])\n    b_selector = ListLastComponentSelector(component_names=["b"])\n    float_round_normalizer = FloatRoundNormalizer(places=1, selectors=[a_selector, b_selector])\n\n    result = differ.diff(OBJ_1, OBJ_2, normalizers=[float_round_normalizer])\n    assert result\n    print(result.support)\n```\n\n<img src="https://github.com/nathan5280/ndl-tools/blob/develop/images/selector-multiple-pass.png" height="75"/>\n\nThere are a few selectors out of the box, but you should subclass your own to minimize the complexity\nof your diff code.\n\n| Selector | Usage |\n| :--- | :--- |\n| ListLastComponentSelector | Match the last component in the element path to a list of names. |\n| ListAnyComponentSelector | Match any component in the element path to a list of names.  Good if you want to select a branch and its child elements. |\n| RegExSelector | Match the element path with the RegEx. |\n| NegativeSelector | Inverts the selection of the Selector it wraps. |\n\n# ListSorters\nListSorters are used to control how lists/sets are sorted.  The are applied using Selectors\nin the same as with Normalizers.  You shouldn\'t need anything other than \nthe two provided ListSorters, but if you need to the extensibility is there.\n',
    'author': 'Nate Atkins',
    'author_email': 'atkinsnw@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nathan5280/ndl-tools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Normalizer used to update a leaf element so that it will compare correctly
with another NDL.  Normalizers are applied to leaf elements using
the Selector associated with the Normalizer.   Normalizers can be chained
so that they are all tried until one succeeds.
"""
import datetime
from abc import abstractmethod
from pathlib import Path
from typing import Any, Optional, List, Union

from .selector import BaseSelector, SELECTORS


class NotNormalizedError(Exception):
    """The normalizer wasn't applied to the element."""


class BaseNormalizer:
    """
    Base normalizer implements the chaining logic.
    """

    def __init__(
        self, selectors: SELECTORS = None,
    ):
        """
        Initialize the normalizer with optional parent normalizer and selector.

        :param selectors: Optional list of selector to use to select which
            elements this normalizer runs.
        """
        if selectors:
            self._selectors = selectors if isinstance(selectors, list) else [selectors]
        else:
            # No selectors specified
            self._selectors = None

    @staticmethod
    def normalize(
        element: Any, path: Path, normalizers: Optional[List["BaseNormalizer"]] = None
    ) -> Any:
        """
        Run all the normalizers until one is applied to normalize the leaf element.

        :param element: Element to normalize.
        :param path: Path to the element.
        :param normalizers: Normalizers to apply to element.
        :return: Normalized element.
        """
        if not normalizers:
            return element

        for normalizer in normalizers:
            if BaseSelector.match(path, normalizer._selectors):
                # Matched drop down and normalize the element.
                try:
                    return normalizer._normalize(element)
                except NotNormalizedError:
                    continue
        return element

    @abstractmethod
    def _normalize(self, element: Any) -> Any:
        """
        Prototype for the core normalizer logic implemented in the subclass.

        :param element: Element to normalize.
        :return: Normalized element.
        """
        pass  # pragma: no cover


NORMALIZERS = Optional[Union[BaseNormalizer, List[BaseNormalizer]]]


class DefaultNormalizer(BaseNormalizer):
    def __init__(
        self, *, selectors: SELECTORS = None,
    ):
        """
        No Op normalizer.

        :param selectors: Optional list of selector to use to select which
            elements this normalizer runs.
        """
        super().__init__(selectors)

    def _normalize(self, element: Any) -> Any:
        """No Op normalizer."""
        return element


# ToDo: Extend this to be smarter about exponential notation
#       and maybe something that adjusts based on the size or places of number being normalize.
class FloatRoundNormalizer(BaseNormalizer):
    def __init__(
        self, places: int, *, selectors: SELECTORS = None,
    ):
        """
        Round a floating point number to a set number of places.

        :param places:  Number of places to round the floating point number to.
        :param selectors: Optional slist of elector to use to select which
            elements this normalizer runs.
        """
        self._places = places
        super().__init__(selectors)

    def _normalize(self, element: Any) -> Any:
        if isinstance(element, float):
            return round(element, self._places)
        raise NotNormalizedError()


class TodayDateNormalizer(BaseNormalizer):
    def __init__(
        self, *, selectors: SELECTORS = None,
    ):
        """
        Convert all dates to today().

        :param selectors: Optional list of selector to use to select which
            elements this normalizer runs.
        """
        super().__init__(selectors)

    def _normalize(self, element: Any) -> Any:
        if isinstance(element, datetime.date):
            return datetime.date.today()
        raise NotNormalizedError


class StrTodayDateNormalizer(BaseNormalizer):
    def __init__(
        self, *, selectors: SELECTORS = None,
    ):
        """
        Overwrite string representation of a date to today().

        :param selectors: Optional list of selector to use to select which
            elements this normalizer runs.
        """
        super().__init__(selectors)

    def _normalize(self, element: Any) -> Any:
        if isinstance(element, str):
            try:
                datetime.date.fromisoformat(element)
                return datetime.date.today().isoformat()
            except ValueError:
                pass
        raise NotNormalizedError()
